/* this is the definition of classes for BFile v0.2
   this package is designed by TAO Yufei (taoyf@cs.ust.hk) 

   this version fixed a bug in v0.1 on the query processing
   a new BTree constructor is added for queries. */
//---------------------------------------------
#ifndef __BFILE
#define __BFILE
//---------------------------------------------
#include "../blockfile/blk_file.h"
#include "../blockfile/cache.h"
#include "../my_math/my_math.h"
#include "../rtree/rtree.h"
//---------------------------------------------
enum BFile_Insrt {BF_NONE, BF_SPLIT};
//---------------------------------------------
class BTree;
class BNode;
class BEntry;
class RTree;
//---------------------------------------------
class BTree
{
public:

	int root;
	BNode *root_ptr;
	RTree *my_rtree;

	//--===functions===--
private:
	void del_root();
public:
	BTree(RTree *_my_rtree);
	BTree(RTree *_my_rtree, int _root);
	~BTree();
	bool insert(int _tm, unsigned char *_bitchar);
	void load_root();
	void agg_qry(bool *_qbitmaps, float *_qmbr, float *_qt, float _ed);
	void get_reg_bitchars(int _tm, unsigned char *_bitchars);
	void load_heap(bool *_qbitmaps, float *_qmbr, float *_qt, float _ed);
};
//---------------------------------------------
class BNode
{
public:
	//information on disk
	int level;
	int num_entries;
	
	//not on disk
	bool dirty;
	int block;
	int node_capacity;  //how many positions can be used
	int page_capacity;  //how many positions in total
	
	BEntry *entries;
	RTree *my_rtree;

	int hashnum;  //number of hash functions
	int bitlen;  //number of bits of each signature
	int bitmaplen;  //number of bits in the bitmap
	int bitcharlen; //number of bytes required to store the bitmaps

	//--===functions===--
private:
	void add_leaf(int _tm, unsigned char *_bitchars);
	int header_size();
	void read_from_buffer(char *_buf);
	void write_to_buffer(char *_buf);
public:
	BNode(RTree *_my_rtree);
	BNode(RTree *_my_rtree, int _blk);
	~BNode();
	void add_new_child(BNode *_new_nd);
	bool chk_ovrflw();
	int choose_subtree(int _key);
	void init_block(RTree *_my_rtree);
	BFile_Insrt insert(int _tm, unsigned char *_bitchar, BNode **_new_nd);
	BFile_Insrt trt_ovrflw(BNode **_new_nd);

	//=====added for the distinct counting problem=====
	bool *get_bitmaps();
	void agg_qry(bool *_qbitmaps, float *_qmbr, float *_qt, int _ed);
	void get_reg_bitchar(int _tm, unsigned char *_bitchars);
};
//---------------------------------------------
class BEntry
{
public:
	//information on disk ----------------------------------------
	int son;
	int tm;
	unsigned char *bitchars;
	//not on disk ------------------------------------------------
	int hashnum;  //number of hash functions
	int bitlen;  //number of bits of each signature
	int bitmaplen;  //number of bits in the bitmap
	int bitcharlen; //number of bytes required to store the bitmaps

	BNode *son_ptr;
	RTree *my_rtree;
	
	//--===functions===--
	BEntry();
	~BEntry();
	void copy(BEntry *_e);
	void del_son();
	bool *get_bitmaps();
	int get_size();
	BNode *get_son();
	void init_entry(RTree *_my_rtree);
	void read_from_buffer(char *_buf);
	void set_to_son(BNode *_succ);
	void write_to_buffer(char *_buf);

	//=====added for the distinct counting problem=====
	void bitmap_or(unsigned char *_bitchars);
	bool agg_qry(bool *_qbitmaps, float *_qmbr, float *_qt, float _ed);
};
//---------------------------------------------
#endif