#include <math.h>
#include <stdio.h>
#include <stdlib.h>
#include "gendef.h"
#include "../my_math/my_math.h"
#include <windows.h>
#include <iostream.h>
#include <tchar.h>
//////////////////////////////////////////////////////////////////////////////
// globals
//////////////////////////////////////////////////////////////////////////////

#ifdef UNIX
void strupr(char *_msg)
{
	int dist = 'A' - 'a';
	char *c_ptr = _msg;

	while (*c_ptr)
	{
		if (*c_ptr >= 'a')
			*c_ptr += dist;
		c_ptr ++;
	}
}
#endif

void error(char *t, bool ex)
{
    fprintf(stderr, t);
    if (ex)
        exit(0);
}

float area(int dimension, float *mbr)
// berechnet Flaeche (Volumen) eines mbr der Dimension dimension
{
    int i;
    float sum;

    sum = 1.0;
    for (i = 0; i < dimension; i++)
	sum *= mbr[2*i+1] - mbr[2*i];

    return sum;
}

float margin(int dimension, float *mbr)
// berechnet Umfang eines mbr der Dimension dimension
{
    float *ml, *mu, *m_last, sum;

    sum = 0.0;
    m_last = mbr + 2*dimension;
    ml = mbr;
    mu = ml + 1;
    while (mu < m_last)
    {
	sum += *mu - *ml;
	ml += 2;
	mu += 2;
    }

    return sum;
}

bool inside(float &p, float &lb, float &ub)
// ist ein Skalar in einem Intervall ?
{
    return (p >= lb && p <= ub);
}

bool inside(float *v, float *mbr, int dimension)
// ist ein Vektor in einer Box ?
{
    int i;

    for (i = 0; i < dimension; i++)
	if (!inside(v[i], mbr[2*i], mbr[2*i+1]))
	    return FALSE;

    return TRUE;
}

// calcutales the overlapping rectangle between r1 and r2
// if rects do not overlap returns null
float* overlapRect(int dimension, float *r1, float *r2)
{
        float *overlap = new float[2*dimension];
        for (int i=0; i<dimension; i++)
        {
            if ((r1[i*2]>r2[i*2+1]) || (r1[i*2+1]<r2[i*2])) // non overlapping
	    {
                delete [] overlap;
		return NULL;
	    }
	    overlap[2*i] = max(r1[i*2], r2[i*2]);
            overlap[2*i+1] = min(r1[i*2+1], r2[i*2+1]);
        }

        return overlap;
}

float overlap(int dimension, float *r1, float *r2)
// calcutales the overlapping area of r1 and r2
// calculate overlap in every dimension and multiplicate the values
{
    float sum;
    float *r1pos, *r2pos, *r1last, r1_lb, r1_ub, r2_lb, r2_ub;

    sum = 1.0;
    r1pos = r1; r2pos = r2;
    r1last = r1 + 2 * dimension;

    while (r1pos < r1last)
    {
	r1_lb = *(r1pos++);
	r1_ub = *(r1pos++);
	r2_lb = *(r2pos++);
	r2_ub = *(r2pos++);

        // calculate overlap in this dimension

        if (inside(r1_ub, r2_lb, r2_ub))
        // upper bound of r1 is inside r2 
	{
            if (inside(r1_lb, r2_lb, r2_ub))
            // and lower bound of r1 is inside
                sum *= (r1_ub - r1_lb);
            else
                sum *= (r1_ub - r2_lb);
	}
	else
	{
            if (inside(r1_lb, r2_lb, r2_ub))
	    // and lower bound of r1 is inside
		sum *= (r2_ub - r1_lb);
	    else 
	    {
		if (inside(r2_lb, r1_lb, r1_ub) &&
		    inside(r2_ub, r1_lb, r1_ub))
	        // r1 contains r2
		    sum *= (r2_ub - r2_lb);
		else
		// r1 and r2 do not overlap
		    sum = 0.0;
	    }
	}
    }

    return sum;
}

void enlarge(int dimension, float **mbr, float *r1, float *r2)
// enlarge r in a way that it contains s
{
    int i;

    *mbr = new float[2*dimension];
    for (i = 0; i < 2*dimension; i += 2)
    {
	(*mbr)[i]   = min(r1[i],   r2[i]);

	(*mbr)[i+1] = max(r1[i+1], r2[i+1]);
    }
}

bool section(int dimension, float *mbr1, float *mbr2)
{
    int i;

    for (i = 0; i < dimension; i++)
    {
	if (mbr1[2*i]     > mbr2[2*i + 1] ||
	    mbr1[2*i + 1] < mbr2[2*i]) 
	    return FALSE;
    }
    return TRUE;
}

bool section_c(int dimension, float *mbr1, float *center, float radius)
{
	float r2;

	r2 = radius * radius;

	if ( (r2 - MINDIST(center,mbr1)) < 1.0e-8)
		return TRUE;
	else
		return FALSE;
	
}

int sort_lower_mbr(const void *d1, const void *d2)
// Vergleichsfunktion fuer qsort, sortiert nach unterem Wert der mbr bzgl.
// der angegebenen Dimension
{
    SortMbr *s1, *s2;
    float erg;
    int dimension;

    s1 = (SortMbr *) d1;
    s2 = (SortMbr *) d2;
    dimension = s1->dimension;
    erg = s1->mbr[2*dimension] - s2->mbr[2*dimension];
    if (erg < 0.0)
	return -1;
    else if (erg == 0.0)
	return 0;
    else 
	return 1;
}

int sort_upper_mbr(const void *d1, const void *d2)
// Vergleichsfunktion fuer qsort, sortiert nach oberem Wert der mbr bzgl.
// der angegebenen Dimension
{
    SortMbr *s1, *s2;
    float erg;
    int dimension;

    s1 = (SortMbr *) d1;
    s2 = (SortMbr *) d2;
    dimension = s1->dimension;
    erg = s1->mbr[2*dimension+1] - s2->mbr[2*dimension+1];
    if (erg < 0.0)
	return -1;
    else if (erg == 0.0)
	return 0;
    else 
	return 1;
}

int sort_center_mbr(const void *d1, const void *d2)
// Vergleichsfunktion fuer qsort, sortiert nach center of mbr 
{
    SortMbr *s1, *s2;
    int i, dimension;
    float d, e1, e2;

    s1 = (SortMbr *) d1;
    s2 = (SortMbr *) d2;
    dimension = s1->dimension;

    e1 = e2 = 0.0;
    for (i = 0; i < dimension; i++)
    {
        d = ((s1->mbr[2*i] + s1->mbr[2*i+1]) / 2.0) - s1->center[i];
        e1 += d*d;
        d = ((s2->mbr[2*i] + s2->mbr[2*i+1]) / 2.0) - s2->center[i];
        e2 += d*d;
    }

    if (e1 < e2)
	return -1;
    else if (e1 == e2)
	return 0;
    else 
	return 1;
}

int sortmindist(const void *element1, const void *element2)
{
    //
    // Vergleichsfunktion fuer die Sortierung der BranchList bei der
    // NearestNarborQuery (Sort, Branch and Prune)
    //

	//The consequence of this sorting is that active branches are sorted
	//in acsending order of their mindist.

    BranchList *e1,*e2;

    e1 = (BranchList *) element1;
    e2 = (BranchList *) element2;
    
    if (e1->mindist < e2->mindist) 
	return(-1);
    else if (e1->mindist > e2->mindist)
	return(1);
    else
	return(0);

}

int pruneBrunchList(float *nearest_distanz, const void *activebrunchList, 
		    int n)
{

    // Schneidet im Array BranchList alle Eintraege ab, deren Distanz groesser
    // ist als die aktuell naeheste Distanz
    //

    BranchList *bl; 
    
    int i,j,k, aktlast;
    
    bl = (BranchList *) activebrunchList;
    
    // 1. Strategie:
    // 
    // Ist MINDIST(P,M1) > MINMAXDIST(P,M2), so kann der 
    // NearestNeighbor auf keinen Fall mehr in M1 liegen!
    //


    aktlast = n;
    

    for( i = 0; i < aktlast ; i++ )
    {
	if (bl[i].minmaxdist < bl[aktlast-1].mindist)
		//bl[aktlast-1] is the maximum mindist of all the branches 
		//if this condition holds, it implies that there exist some branches whose
		//mindists are greater than minmandist of this branch. in the sequel, all these
		//branches will be removed from further consideration. 
	    for( j = 0; (j < aktlast) ; j++ )
		if ((i!=j) && (bl[j].mindist>bl[i].minmaxdist))
		{
		    aktlast = j;

			//checkpoint
			if (j <= i)
				error("Error in pruneBrunchList\n", true);
		    break;
		}
    }

    

    // 2. Strategie:
    //
    // nearest_distanz > MINMAXDIST(P,M)
    // -> nearest_distanz = MIMMAXDIST(P,M)
    //


    for( i = 0; i < aktlast; i++)
	if (*nearest_distanz > bl[i].minmaxdist)
	    *nearest_distanz = bl[i].minmaxdist;

    
    // 3. Strategie:
    // 
    // nearest_distanz < MINDIST(P,M)
    //
    // in M kann der Nearest-Narbor sicher nicht mehr liegen.
    //

    for( i = 0; (i < aktlast) && (*nearest_distanz >= bl[i].mindist) ; i++);
    
    aktlast = i;

    // printf("n: %d aktlast: %d \n",n,aktlast);   

    return (aktlast);
   
}

float objectDIST(float *p1, float *p2)
{

    //
    // Berechnet das Quadrat der euklid'schen Metrik.
    // (Der tatsaechliche Abstand ist uninteressant, weil
    // die anderen Metriken (MINDIST und MINMAXDIST fuer
    // die NearestNarborQuery nur relativ nie absolut 
    // gebraucht werden; nur Aussagen "<" oder ">" sind
    // relevant.
    //

    float summe = 0;
    int i;
    
    for( i = 0; i < DIMENSION; i++)
	summe += pow(p1[i] - p2[i], 2 );

    //return( sqrt(summe) );
    return(summe);
}

float MINDIST(float *p, float *bounces)
{
    //
    // Berechne die kuerzeste Entfernung zwischen einem Punkt Point
    // und einem MBR bounces (Lotrecht!)
    //

    float summe = 0.0;
    float r;
    int i;

    for(i = 0; i < DIMENSION; i++)
    {
	if (p[i] < bounces[2*i])
	    r = bounces[2*i];
	else
	{
	    if (p[i] > bounces[2*i +1])
		r = bounces[2*i+1];
	    else 
		r = p[i];
	}    

	summe += pow( p[i] - r , 2);
    }
    return(summe);
    
}

float MINMAXDIST(float *p, float *bounces)
{

    // Berechne den kleinsten maximalen Abstand von einem Punkt Point
    // zu einem MBR bounces.
    // Wird benutzt zur Abschaetzung von Abstaenden bei NearestNarborQuery.
    // Kann als Supremum fuer die aktuell kuerzeste Distanz: 
    // Alle Punkte mit einem Abstand > MINMAXDIST sind keine Kandidaten mehr
    // fuer den NearestNarbor
    // vgl. Literatur: 
    // Nearest Narbor Query v. Roussopoulos, Kelley und Vincent, 
    // University of Maryland
    
    float summe = 0;
    float minimum = 1.0e20;
    float S = 0;

    float rmk, rMi;
    int k,i;

    for( i = 0; i < DIMENSION; i++) 
    { 
	rMi = (	p[i] >= (bounces[2*i]+bounces[2*i+1])/2 )
	    ? bounces[2*i] : bounces[2*i+1];
	S += pow( p[i] - rMi , 2 );
    }

    for( k = 0; k < DIMENSION; k++)
    {  
	
	rmk = ( p[k] <=  (bounces[2*k]+bounces[2*k+1]) / 2 ) ? 
	    bounces[2*k] : bounces[2*k+1];

	summe = pow( p[k] - rmk , 2 );	
	
	rMi = (	p[k] >= (bounces[2*k]+bounces[2*k+1]) / 2 )
	    ? bounces[2*k] : bounces[2*k+1];

	summe += S - pow( p[k] - rMi , 2 );
	
	minimum = min( minimum,summe);
    }

    return(minimum);
    

}

bool contain(float *_m1, float *_m2, int dim)
//check if _m1 contains _m2
{
	for (int i = 0; i < dim; i ++)
	{
		if (!(_m1[2 * i] <= _m2[2 * i] && 
			  _m2[2 * i + 1] <= _m1[2 * i + 1]))
			return false;
	}
	return true;	
}

bool equal(float *_m1, float *_m2, int dim)
//check if the two mbrs are equal
{
	for (int i = 0; i < 2 * dim; i ++)
		if (_m1[i] - _m2[i] > FLOATZERO)
			return false;
	return true;
}

/*****************************************************************
this function transforms a bitmap to a char array suitable for storage
bitmap: the bitmap to be converted
charmap: the char array to be returned
len: the number of bits in bitmap
if all bits are false then this funciton return true;
Coded by yufei tao may 2003
*****************************************************************/
bool bitmap2char(bool *_bitmap, unsigned char *_charmap, int _len, int _clen)
{
	bool allfalse=true;
	for (int i=0; i<_clen; i++)
		_charmap[i]=0;
	for (i=0; i<_len; i++)
	{
		if (_bitmap[i])
		{
			allfalse=false;
			int sub=i/8, offset=7-i%8;
			unsigned char tmp=1;
			tmp=tmp<<offset;
			_charmap[sub]|=tmp;
		}
	}
	return allfalse;
}

/*****************************************************************
this function is the reverse of bitmap2char
bitmap: the bitmap to be returned
charmap: the char array to be returned
len: the number of bits in bitmap
if all bits are false then this funciton return true;
Coded by yufei tao may 2003
*****************************************************************/
void char2bitmap(bool *_bitmap, unsigned char *_charmap, int _len)
{
	for (int i=0; i<_len; i++)
		_bitmap[i]=false;
	for (i=0; i<_len; i++)
	{
		int sub=i/8, offset=7-i%8;
		if ((_charmap[sub]>>offset)%2)
			_bitmap[i]=true;
	}
}

/*****************************************************************
this function transforms a bitmap to a char array suitable for storage
using compression techniques using ANF and group by 2, Compression ratio = 0.31
bitmap: the bitmap to be converted
charmap: the char array to be returned, including the aray for the leading 1's ending position and trailing 0's position, each 4 bits, 
for each group to be returned. that's where we start convert to charmap
len: the number of bits in each single bitmap (Note: not the total number of bits in all bitmaps, total= len*hashnum)
hashnum: the number of hash functions (i.e. number of bitmaps), it must be an even number
coded by feifei dec 2003
*****************************************************************/
void compression_bitmap2char(bool *_bitmap, unsigned char *_charmap, int _len, int _hashnum, int _clen, int& used_clen)
{
	int i, j,onecount=0,zerocount=0,remaining_bits,start=0,end=0;
	unsigned int key1=0,key0=0;
	int total_len = _len*_hashnum;
	int conca_len = _len*2;
	int num_groups = _hashnum/2;
	int groupNum=0;
	int clen_counter=0;
	used_clen=0;
	end = start+2;
	for (i=0; i<_clen; i++)
		_charmap[i]=0;
	while(end<_hashnum)
	{
		bool* concatenate_bitmap=new bool[conca_len];
		int offset1 = _len*start;
		for (i=0;i<_len;i++)
		{
			for (j=start;j<end;j++)
				concatenate_bitmap[j-start+i*2] = _bitmap[offset1+(j-start)*_len+i];			
		}
		for (i=0; i<conca_len; i++)
		{	
			if (concatenate_bitmap[i])
				onecount++;
			else
				break;
		}
		for (i=conca_len-1;i>=0;i--)
		{
			if (!concatenate_bitmap[i])
				zerocount++;
			else
				break;
		}
		remaining_bits = total_len-onecount-zerocount;
		if (remaining_bits%8==0)
		{
			key1= onecount;
			key0= conca_len-zerocount;
		}
		else
		{
			key1 = onecount;
			int tmp = remaining_bits/8;
			int padding = remaining_bits-8*tmp;
			key0 = conca_len-zerocount+padding; 
		}
		unsigned int key0padding = conca_len-key0;
		unsigned char location = (unsigned char)(((key1 & 0xF)<<4) | (key0padding & 0xF));
		location = location & 0xFF;
		_charmap[used_clen] = location;
		for (i=key1;i<key0;i++)
		{
			if ( (i-key1)%8==0 )
				used_clen++;
			_charmap[used_clen] = 0;
			if (concatenate_bitmap[i])
			{	
				unsigned char tmp=1;
				int offset = 7-(i-key1)%8;
				_charmap[used_clen]= tmp<<offset | _charmap[used_clen]; 
			}
			clen_counter++;
		}
		used_clen++;
		start=end;
		end = start+2;		
	}
}


/*****************************************************************
this function is the reverse of compression_bitmap2char
bitmap: the bitmap to be returned
charmap: the char array for input
len: the number of bits in each bitmap
hashnum: the number of bitmaps, i.e. number of hash funcions
clen: the length of the _charmap
Coded by feifei dec 2003
*****************************************************************/
void compression_char2bitmap(bool *_bitmap, unsigned char *_charmap, int _len, int _hashnum, int _clen)
{
	int counter = 0;
	int group = 0;
	int i,j;
	int total_len = _len*_hashnum;
	for (i=0;i<total_len;i++)
		_bitmap[i]=false;
	while (counter < _clen)
	{
		unsigned char location = _charmap[counter];
		unsigned int key1 = (location & 0xFF) >>4;
		unsigned int key0 = (location & 0xF);
		bool* concatenate_bitmap =  new bool[_len*2];
		for (i=0;i<_len*2;i++)
			concatenate_bitmap[i]=false;
		for (i=0;i<key1;i++)
			concatenate_bitmap[i]=true;
		int remainingbits  = _len*2-key1-key0;
		int length = remainingbits/8;
		int cc=0;
		for (i=counter+1;i<=counter+length;i++)
		{
			for (j=0;j<8;j++)
			{	
				int offset = 7-j%8;
				if ((_charmap[i]>>offset)& 0x1)
					concatenate_bitmap[key1+cc]=true;
				cc++;
			}
		}
		int base=group*2*_len;
		for (int i=0;i<2*_len;i++)
		{
			if (concatenate_bitmap[i]==true)
			{
				if (i%2==0)
					_bitmap[base+i/2]=true;
				else
					_bitmap[base+_len+(i-1)/2]=true;
			}
		}
		counter=counter+length+1;
		group++;
	}
}

/*****************************************************************
this function returns the total number of ones in the bitchar
para
bitchars: the input bitchar
len: number of bits in the bitmap
Coded by yufei tao may 2003
*****************************************************************/

int get_one_number(unsigned char *_bitchars, int _len)
{
	bool *bitmaps=new bool[_len];
	char2bitmap(bitmaps, _bitchars, _len);
	int cnt=0;
	for (int i=0; i<_len; i++)
		if (bitmaps[i])
			cnt++;
	delete []bitmaps;
	if (cnt>1)
		printf("");
	return -cnt;
}

/*****************************************************************
this function writes an integer into a file opened in the binary mode
Coded by yufei tao may 2003
*****************************************************************/
void binary_int_write(FILE *_fp, int _n)
{
	char *c=(char *)&_n;
	for(int i=0; i<sizeof(int); i++)
	{
		fputc(*c, _fp); c++;
	}
}

/*****************************************************************
this function reads an integer From a file opened in the binary mode
Coded by yufei tao may 2003
*****************************************************************/
void binary_int_read(FILE *_fp, int *_n)
{
	char *c=(char *)_n;
	for(int i=0; i<sizeof(int); i++)
	{
		*c=fgetc(_fp); c++;
	}
}