/* rtree.h
   this file defines the class RTree*/

#ifndef __RTREE
#define __RTREE
//------------------------------------------------------------
#include "../func/gendef.h"
#include "../bfile/bfile.h"
#include "../heap/heap.h"
//------------------------------------------------------------
class LinList;
class SortedLinList;
class Cache;
class RTNode;
class Entry;
//------------------------------------------------------------
class RTree : public Cacheable
{
public:
	bool *re_level;                      // if re_level[i] is TRUE,
							             // there was a reinsert on level i
	bool root_is_data;                   // TRUE, if root is a data page
    int dimension;                       // dimension of the data's
	int fin_time;                        // the final timestamp indexed
    int num_of_data;	                 // # of stored data
    int num_of_dnodes;	                 // # of stored data pages
    int num_of_inodes;	                 // # of stored directory pages
	int root;                            // block # of root node

	RTNode *root_ptr;                    // root-node
    LinList *re_data_cands;               // data entries to reinsert

	int hashnum;  //number of hash functions
	int bitlen;  //number of bits of each signature
	int bitmaplen;  //number of bits in the bitmap
	int bitcharlen; //number of bytes required to store the bitmaps
	bool pcsa; //toggle whether pcsa is used in estimation

	Heap *hp;

	//for debugging purpose
//	FILE *qfp;
	//==========================================================
    RTree(char *_fname, Cache *_cache);
    RTree(char *_dfname, char *_tfname, int _b_len, Cache *_cache, int _dim, int _hashnum, int _bitlen,
		  bool _pcsa);
    ~RTree();

	void del_root();
	bool findLeaf(Entry *e);
	int get_num()                        // returns # of stored data
		{ return num_of_data; }
   	void load_root();                    // loads root_node into memory
	void read_header(char *buffer);      // reads Rtree header
	void write_header(char *buffer);     // writes Rtree header
    void insert(Entry *d);               // inserts new data into tree
    void rangeQuery(float *mbr, SortedLinList *res);
	//=====added for arb-trees====================================
	void create_btrees(char *_dfname, char *_sigfname, int _rnum);
	void insert_bitmap(int _tm, int _rid, float *_rmbr, unsigned char *_bitchar);
	float agg_qry(float *_qmbr, float *_qt, float _ed);
	float agg_qry2(float *_qmbr, float *_qt, float _ed);
	void get_reg_bitchars(int _rid, float *_rmbr, int _tm, unsigned char *_bitchars);
	void load_heap(bool *_qbitmaps, float *_qmbr, float *_qt, float _ed);
	bool visit_child(bool *_qbitmaps, char *_bitchars);
	void conservative_pos_est(bool *_qbitmaps, Heap *_hp, int *_pos);
	bool visit_child(bool *_qbitmaps, unsigned char *_bitchars, int *_pos);
};

#endif // __RTREE
